type t = Mustache.t

module Configuration = struct
  type path_t = Fpath.t option

  type paths_t = {
      dir    : path_t;
      header : path_t;
      note   : path_t;
      front  : path_t;
      list   : path_t;
      item   : path_t;
    }

  let of_config config =
    let open Confix.ConfixToml in
    let p k = path config ("templates" / k) in
    {
      dir    = p "dir";
      header = p "header";
      note   = p "note";
      front  = p "front";
      list   = p "list";
      item   = p "item";
    }
end

type header_t = Header of t
type footer_t = Footer of t
type list_t   = List of t
type item_t   = Item of t
type note_t   = Note of t
type front_t  = Front of t

let of_string = Mustache.of_string
let of_file f = File.load f |> of_string

let map_tpl_opt f field ps =
  let open Configuration in
  match ps.dir, field with
  | Some dir, Some bn -> Some (f (of_file (Fpath.append dir bn)))
  | _ -> None

let string s = [Html.data s]
let section ~inverted name contents = [Html.unescaped_data ("section " ^ String.concat "." name)]
let unescaped elts = [Html.unescaped_data "use escaped instead"]
let partial ?indent name _ _ = [Html.data "partials not supported"]
let comment _ = [Html.data ""]
let concat = List.concat

let escaped_index ~from ~n metas e = [Html.data "temp"]
  (* match List.hd e with *)
  (* | "topics" -> *)
  (*    let topics = *)
  (*      ListLabels.fold_left *)
  (*        ~init:(Logarion.Meta.StringSet.empty) *)
  (*        ~f:(fun a e -> Logarion.Meta.unique_topics a e ) metas *)
  (*    in *)
  (*    Logarion.Meta.StringSet.fold (fun e a -> a ^ "<li><a href=\"/topic/" ^ e ^ "\">" ^ e ^ "</a></li>") topics "" *)

let header_fold template archive =
  Mustache.fold ~string ~section ~escaped:(Html.Renderer.archive archive) ~unescaped ~partial ~comment ~concat template

let header archive =
  Html.(header [title [anchor "index.html" [data archive.Logarion.Archive.Configuration.title]]])

let note_fold template note =
  Mustache.fold ~string ~section ~escaped:(Html.Renderer.note note) ~unescaped ~partial ~comment ~concat template

let meta meta =
  let open Logarion.Note in
  let open Logarion.Meta in
  let abstract = meta.abstract in
  let author = meta.author.name in
  let date = Date.(pretty_date @@ last meta.date) in
  let series = stringset_csv meta.series in
  let topics = stringset_csv meta.topics in
  let keywords = stringset_csv meta.keywords in
  let uuid = Id.to_string meta.uuid in
  Html.meta ~abstract ~author ~date ~series ~topics ~keywords ~uuid

let body note =
  Html.note
    [ Html.title [Html.unescaped_data note.Logarion.Note.meta.Logarion.Meta.title]; (* Don't add title if body contains one *)
      meta note.meta;
      Html.unescaped_data @@ Omd.to_html @@ Omd.of_string note.Logarion.Note.body ]

let listing metas = Html.listing metas

let page ~style title header body = Html.to_string @@ Html.page ~style title header body
